// Title:  BiDiPro DLL module
// Desc:   16-bit 5V-only flashes
// Ver:    0.0.3
// Found:  2007-08-22
// Last:   2009-06-15
// CPU:    386+
// Model:  Tiny
// Compil: Borland C++ 3.1 (DOS)
// Author: Oleg O. Chukaev

// Based on Atmel, Winbond and Sanyo datasheets


#include <dos.h>
#include <conio.h>
#include <stdlib.h>
#include "..\dlldata.h"
#include "..\errors.h"
#include "..\bidi.h"

#define dlp ((DLLDATA *) NULL)        /* ⥫    */

#define bdport dlp->bdprt            /* 㪠⥫  hardware ࠩ */

#define DataPortIn()  bdport(PFORCE|PWR1|PAIN|PBIN)
#define DataPortOut() bdport(PFORCE|PWR1)

#define ATMP 64


struct  CHDEFS {
	dword size;   // 
	byte  dpins;  //   dip 
};

CHDEFS chd[]={
//   size    pins
  {0x008000L, 32},  //256
  {0x010000L, 32},  //516
  {0x020000L, 32},  //1024
  {0x040000L, 32},  //2048
  {0x080000L, 32},  //4096
  {0x100000L, 32},  //8192
};


// ७   CFG:
static struct CFGKEY cfgkeys[] = {
{'D', "Vccp",  50,50,70},      //殮 ⠭  ணࠬ஢   0
{'D', "Vccv",  50,50,70},      //殮 ⠭  䨪樨        1
{'D', "ErTime",10000,10000,0}, //p 横 p (ms)                 2
{'D', "WrTime",50,50,0},       //p  ᫮ (mcs)                  3
{'d', "Nnum",  25,25,0},       //᫮ ⮪ ணࠬ஢            4
{'D', "IdAlg", 0,0,0},         // ⥭ ID                        5
{'D', "ErAlg", 0,0,0},         // ࠭                         6
{'h', "sec_0", 0,0,0},         //ᠭ ᥪ஢                         7
{'h', "sec_1", 0,0,0},         //ᠭ ᥪ஢                         8
{'h', "sec_2", 0,0,0},         //ᠭ ᥪ஢                         9
{'h', "sec_3", 0,0,0},         //ᠭ ᥪ஢                        10
{'D', "WrAlg", 0,0,0},         //                           11
{'d', "test",  0,0,0},         // 祭 -              12
{0,    NULL,   0,0},           //End of table
};

#define cfg_Vccp   cfgkeys[0].data
#define cfg_Vccv   cfgkeys[1].data
#define cfg_ErTime cfgkeys[2].data
#define cfg_WrTime cfgkeys[3].data
#define cfg_Nnum   cfgkeys[4].data
#define cfg_IdAlg  cfgkeys[5].data
#define cfg_ErAlg  cfgkeys[6].data
#define cfg_sec_0  cfgkeys[7].data
#define cfg_sec_1  cfgkeys[8].data
#define cfg_sec_2  cfgkeys[9].data
#define cfg_sec_3  cfgkeys[10].data
#define cfg_WrAlg  cfgkeys[11].data
#define cfg_test   cfgkeys[12].data

static char *copyright = "16-bit Flash module (c) Oleg O. Chukaev ["__DATE__"]";
static word chtype = 0xFFFF;           //  ROM
static CHDEFS *pchd;

#define power_off breset


void breset (void) {
	bdport (PRST);
}


void dbgBox (char *str) {
	dlp->msgBox (dmfError | dmfOKButton, str);
}


int power_on (byte mode) {

	word vmask = 0x80;
	word vpp = 55;
	word vid = 140;
	/* Vid   ⠭  14V  */
	/* 祭 ⠡쭮 0  A9 */

	if (chtype == 0xFFFF)
		return (ERC_CHIPUNSP);

	if (mode == 'R') {
		DataPortIn ();			/* DD1:IIO */
		bdport (PFORCE | PWR2);		/* DD5:OOO */
		bdport (PFORCE | PWA2 | BITTER | BITA18 | BITA19);
		dlp->voltage (dlp->Vcc, vpp, vid);
		bdport (PWB2 | vmask);
	} else if (mode == 'V') {
		DataPortIn ();			/* DD1:IIO */
		bdport (PFORCE | PWR2);		/* DD5:OOO */
		bdport (PFORCE | PWA2 | BITTER | BITA18 | BITA19);
		dlp->voltage (cfg_Vccv, vpp, vid);
		bdport (PWB2 | vmask);
	} else if (mode == 'W') {
		bdport (PFORCE | PWR2);		/* DD5:OOO */
		bdport (PFORCE | PWA2 | BITTER | BITA18 | BITA19);
		DataPortOut ();			/* DD1:OOO */
		dlp->voltage (cfg_Vccp, vpp, vid);
		bdport (PWB2 | vmask );		/* Vcc on */
	} else if (mode == 'I') {
		bdport (PFORCE | PWR2);
		bdport (PFORCE | PWA2 | BITTER | BITA18 | BITA19);
		DataPortOut ();
		vid = dlp->Vid;
		dlp->voltage (dlp->Vcc, vpp, vid);
		if (dlp->Vid > 0)
			vmask |= 0x10;		/* Vid enable */
		bdport (PWB2 | vmask );		/* Vcc, Vid on */
	}
	dlp->mcsdelay (100000L);
	return (0);
}


void write_code16 (dword addr, word data) {
	bdport (PWIR | (byte)((addr >> 0) & 0xff));	/* A0..A7 */
	bdport (PWC1 | (byte)((addr >> 8) & 0xff));	/* A8..A15 */
	bdport (PWA2 | BITTER | BITOE | BITA18 | BITA19
		| (byte)((addr >> 16) & 0x03));		/* A16..A17 */
	bdport (PWA1 | (data & 0xff));
	bdport (PWB1 | (data >> 8));
	bdport (PWA2 | BITTER | BITOE | BITA19 | (byte)((addr >> 16) & 0x03));
	dlp->mcsdelay (2);
	bdport (PWA2 | BITTER | BITOE | BITA18 | BITA19
		| (byte)((addr >> 16) & 0x03));
}


/* ⥭ ᫮. adr  ᮬ _᫮_ */
inline word read_one (dword adr) {

	/* A0..A7 */
	bdport (PWIR | (byte)((adr >> 0) & 0xff));
	/* A8..A15 */
	bdport (PWC1 | (byte)((adr >> 8) & 0xff));
	/* A16..A17, -CE == 0, -OE == 0, -WE == 1, -RESET == 1 */
	bdport (PWA2 | BITTER | BITA18 | BITA19 | (byte)((adr >> 16) & 0x03));
	/* ⠥  */
	return (((byte)bdport(PRB1) << 8) + (byte)bdport(PRA1));
}


/*  ᫮. adr  ᮬ _᫮_ */
void write_one (dword adr, word dat) {
	write_code16 (0x5555, 0xaa);
	write_code16 (0x2aaa, 0x55);
	write_code16 (0x5555, 0xa0);
	write_code16 (adr, dat);
}


/* ⨥     䫥襪  Sanyo */
inline void sanyo_unlock (void) {
	read_one (0x1823);
	bdport (PWA2 | BITOE | BITCE | BITTER | BITA18 | BITA19);
	read_one (0x1820);
	bdport (PWA2 | BITOE | BITCE | BITTER | BITA18 | BITA19);
	read_one (0x1822);
	bdport (PWA2 | BITOE | BITCE | BITTER | BITA18 | BITA19);
	read_one (0x0418);
	bdport (PWA2 | BITOE | BITCE | BITTER | BITA18 | BITA19);
	read_one (0x041b);
	bdport (PWA2 | BITOE | BITCE | BITTER | BITA18 | BITA19);
	read_one (0x0419);
	bdport (PWA2 | BITOE | BITCE | BITTER | BITA18 | BITA19);
	read_one (0x041a);
	bdport (PWA2 | BITOE | BITCE | BITTER | BITA18 | BITA19);
}


/*  ᥣ 稯, Atmel */
int atmel_write (void) {

	dword adr;
	word n, t;
	dword istep = pchd->size / 64L;
	word tdr, tdw;

	power_on ('W');
	for (adr = 0; adr < pchd->size; adr += 2) {
		tdw = dlp->getbytebuf (adr) + (dlp->getbytebuf (adr+1) << 8);
		for (n = 0; n <= cfg_Nnum; n++) {
			DataPortOut ();
			write_one (adr >> 1, tdw);

			DataPortIn ();
			for (t = 0; t < cfg_WrTime; t++) {
				dlp->mcsdelay (1);
				tdr = read_one (adr >> 1);
				if ((tdr & 0x80) == (tdw & 0x80))
					break; //right data
			}

			if (tdr == tdw)
				break;
			if (n == cfg_Nnum) {   // error
				dlp->adr = adr;
				dlp->edata = tdr;
				power_off ();
				return (ERC_WRITE);
			}
		}
		if (adr % istep == 0 && dlp->iprogress (adr >> 1, 2))
			break;
	}
	power_off ();
        return (0);
}


/*  ᥣ 稯, Sanyo */
/* : ஢ 砭   - toggle bit  data polling */
/* : ஢ ⢠  */
int sanyo_write (void) {

	dword adr;
	dword istep = pchd->size / 64L;
	word tdw;

	power_on ('R');
	DataPortIn ();
	sanyo_unlock ();
	DataPortOut ();
	for (adr = 0; adr < pchd->size; adr += 2) {
		tdw = dlp->getbytebuf (adr) + (dlp->getbytebuf (adr+1) << 8);
		write_code16 (0, 0x0010);
		write_code16 (adr >> 1, tdw);
		dlp->mcsdelay (cfg_WrTime);
		if (adr % istep == 0 && dlp->iprogress (adr >> 1, 2))
			break;
	}
	power_off ();
        return (0);
}


/*  ࠭, Winbond */
/* : ஢ 砭   - toggle bit  data polling */
int winbond_write_page (dword addr) {

	word tdw, p_size;
	int i;

	p_size = cfg_sec_0 & 0xffff;

	for (i = 0; i < p_size; i++) {
		tdw = dlp->getbytebuf ((addr+i)*2) + (dlp->getbytebuf ((addr+i)*2+1) << 8);
		bdport (PWIR | (byte)((addr+i >> 0) & 0xff));	/* A0..A7 */
		bdport (PWC1 | (byte)((addr+i >> 8) & 0xff));	/* A8..A15 */
		bdport (PWA2 | BITTER | BITOE | BITA19		/* WE = 0 */
			| (byte)((addr+i >> 16) & 0x03));	/* A16..A17 */
		bdport (PWA1 | (tdw & 0xff));
		bdport (PWB1 | (tdw >> 8));
		bdport (PWA2 | BITTER | BITOE | BITA18 | BITA19
			| (byte)((addr+i >> 16) & 0x03));
	}
	dlp->mcsdelay (cfg_WrTime);
	return (0);
}


/*  ᥣ 稯, Winbond */
int winbond_write (void) {

	dword adr;
	dword istep = pchd->size / 64L;
	word p_size;

	if (chtype == 0xFFFF)
		return (ERC_CHIPUNSP);

	p_size = cfg_sec_0 & 0xffff;

	power_on ('W');
	DataPortOut ();			// DD1:OIO
	write_code16 (0x5555, 0xaaaa);
	write_code16 (0x2aaa, 0x5555);
	write_code16 (0x5555, 0x8080);
	write_code16 (0x5555, 0xaaaa);
	write_code16 (0x2aaa, 0x5555);
	write_code16 (0x5555, 0x2020);
	dlp->mcsdelay (10000);
	bdport (PWA2 | BITTER | BITOE | BITA18);
	for (adr = 0; adr < pchd->size / 2; adr += p_size) {
		winbond_write_page (adr);
		if (adr % istep == 0 && dlp->iprogress (adr, 2))
			break;
	}
	power_off ();
	return (0);
}


/* ࠭ 稯, Atmel */
int atmel_erase (void) {

	dword j;
	word tdr = 0;
	int i, er_ok = 0;

	dlp->iprogress (ATMP, 0);
	power_on ('W');

	for (i = 1; i <= ATMP; i++) {
		DataPortOut ();
		write_code16 (0x5555, 0xaa);
		write_code16 (0x2aaa, 0x55);
		write_code16 (0x5555, 0x80);
		write_code16 (0x5555, 0xaa);
		write_code16 (0x2aaa, 0x55);
		write_code16 (0x5555, 0x10);

		if (dlp->iprogress (i, 2))
			break;  //progress

		DataPortIn ();
		bdport (PWA2 | BITTER | BITA18 | BITA19);
		tdr = read_one (0) & 0x40;
		bdport (PWA2 | BITTER | BITA18 | BITA19 | BITOE);
		for (j = 0; j < cfg_ErTime; j++) {
			dlp->mcsdelay (1000);
			bdport (PWA2 | BITTER | BITA18 | BITA19);
			if ((read_one (0) & 0x40) == tdr) {
				er_ok = 1;
				break;
			} else {
				tdr ^= 0x40;
			}
			bdport (PWA2 | BITTER | BITA18 | BITA19 | BITOE);
		}
		if (er_ok)
			break;
	}
	dlp->iprogress (0, 1); //close
	power_off ();
        return (0);
}


/* ࠭ 稯, Sanyo */
/* : ஢    । ࠭ */
int sanyo_erase (void) {

        word s_num, s_size;
	dword j;
	int i;

	s_num = cfg_sec_0 >> 16;
	s_size = cfg_sec_0 & 0xffff;

	power_on ('R');
	DataPortIn ();
	sanyo_unlock ();

	dlp->iprogress (s_num, 0);

	for (i = 0; i < s_num; i++) {
		DataPortOut ();
		write_code16 (0, 0x0020);
		write_code16 (i * s_size, 0x00d0);

		if (dlp->iprogress (i, 2))
			break;  //progress

		for (j = 0; j < cfg_ErTime; j++) {
			dlp->mcsdelay (1000);
		}
	}
	dlp->iprogress (0, 1); //close
	power_off ();
        return (0);
}


/* ࠭ 稯, Winbond */
int winbond_erase (void) {

	if (chtype == 0xFFFF)
		return (ERC_CHIPUNSP);

	power_on ('W');
	DataPortOut ();
	write_code16 (0x5555, 0xaaaa);
	write_code16 (0x2aaa, 0x5555);
	write_code16 (0x5555, 0x8080);
	write_code16 (0x5555, 0xaaaa);
	write_code16 (0x2aaa, 0x5555);
	write_code16 (0x5555, 0x1010);
	dlp->mcsdelay (cfg_ErTime * 1000L);
	power_off ();

	return (0);
}


/* ⥭ ᥣ 稯. */
int read_all (void) {

	dword adr;
	dword istep = pchd->size / 64L;
	word dat;

	if (chtype == 0xFFFF)
		return (ERC_CHIPUNSP);

	power_on ('R');
	for (adr = 0; adr < pchd->size; adr += 2) {
		dat = read_one (adr >> 1);
		dlp->setbytebuf (adr, dat & 0xff);
		dlp->setbytebuf (adr + 1, dat >> 8);
		if (adr % istep == 0 && dlp->iprogress (adr >> 1, 2))
			break;
	}
	power_off ();
	return (0);
}


/*  ᥣ 稯. */
int write_all (void) {

	int rc;

	if (chtype == 0xFFFF)
		return (ERC_CHIPUNSP);

        if (cfg_WrAlg == 0) {
        	/* ᫮ ,   Atmel */
                rc = atmel_write ();
        } else if (cfg_WrAlg == 1) {
        	/* ᫮ ,   Sanyo */
                rc = sanyo_write ();
        } else if (cfg_WrAlg == 2) {
        	/* ࠭筠 ,   Winbond */
		rc = winbond_write ();
        }

	return (rc);
}


/* ࠭ 稯. */
int erase (void) {

        int rc;

	if (chtype == 0xFFFF)
		return (ERC_CHIPUNSP);

        if (cfg_ErAlg == 0) {
                rc = atmel_erase ();
        } else if (cfg_ErAlg == 1) {
        	rc = sanyo_erase ();
        } else if (cfg_ErAlg == 2) {
        	rc = winbond_erase ();
        }

	return (rc);
}


/* ⥭ 䨪. */
int read_id (void) {

	dword id0, id1, id2, id3;
	dlp->edata = 0;

	if (chtype == 0xFFFF)
		return (ERC_CHIPUNSP);

	power_on ('I');

	if (cfg_IdAlg == 0) {
		/* Intelligent (software) ID */
		DataPortOut ();
		write_code16 (0x5555, 0xaa);
		write_code16 (0x2aaa, 0x55);
		write_code16 (0x5555, 0x90);
		bdport (PWA2 | BITTER | BITA18 | BITA19);

		DataPortIn ();			/* DD1:IIO */
		bdport (PWC1 | 0);		/* 2   */
		bdport (PWIR | 0);		/* 1   */
		id0 = (byte)bdport (PRA1);	/* Mfr code */
		id1 = (byte)bdport (PRB1);	/* Mfr code */
		bdport (PWIR | 1);		/* 1   = 1 */
		id2 = (byte)bdport (PRA1);	/* Dev code */
		id3 = (byte)bdport (PRB1);	/* Dev code */

		DataPortOut ();
		write_code16 (0, 0xf0);
	} else if (cfg_IdAlg == 1) {
		/* Hardware ID */
		bdport (PWA2 | BITTER | BITA18 | BITA19 | BITPE);
		dlp->mcsdelay (10);

		DataPortIn ();			/* DD1:IIO */
		bdport (PWC1 | 0);		/* 2   */
		bdport (PWIR | 0);		/* 1   = 0 */
		id0 = (byte)bdport (PRA1);	/* Mfr code */
		id1 = (byte)bdport (PRB1);	/* Mfr code */
		bdport (PWIR | 1);		/* 1   = 1 */
		id2 = (byte)bdport (PRA1);	/* Dev code */
		id3 = (byte)bdport (PRB1);	/* Dev code */

		bdport (PWA2 | BITTER | BITA18 | BITA19);
	} else if (cfg_IdAlg == 2) {
		/* Intelligent (software) ID, Winbond */
		DataPortOut ();
		write_code16 (0x5555, 0xaaaa);
		write_code16 (0x2aaa, 0x5555);
		write_code16 (0x5555, 0x9090);
		bdport (PWA2 | BITTER | BITA18 | BITA19);
		dlp->mcsdelay (10);

		DataPortIn ();			/* DD1:IIO */
		bdport (PWC1 | 0);		/* 2   */
		bdport (PWIR | 0);		/* 1   */
		id0 = (byte)bdport (PRA1);	/* Mfr code */
		id1 = (byte)bdport (PRB1);	/* Mfr code */
		bdport (PWIR | 1);		/* 1   = 1 */
		id2 = (byte)bdport (PRA1);	/* Dev code */
		id3 = (byte)bdport (PRB1);	/* Dev code */

		DataPortOut ();
		write_code16 (0x5555, 0xaaaa);
		write_code16 (0x2aaa, 0x5555);
		write_code16 (0x5555, 0xf0f0);
		bdport (PWA2 | BITTER | BITA18 | BITA19);
		dlp->mcsdelay (10);
	}

	dlp->edata = id0 | (id1 << 8) | (id2 << 16L) | (id3 << 24L);
	power_off ();
	return 0;
}


/*  祭 -,  Atmel */
int test (void) {

	dword id;

	if (chtype == 0xFFFF)
		return (ERC_CHIPUNSP);

	power_on ('I');

	/* Intelligent (software) ID */
	DataPortOut ();
	write_code16 (0x5555, 0xaa);
	write_code16 (0x2aaa, 0x55);
	write_code16 (0x5555, 0x90);
	bdport (PWA2 | BITTER | BITA18 | BITA19);

	DataPortIn ();			/* DD1:IIO */
	bdport (PWC1 | 0);		/* 2   */
	bdport (PWIR | 2);		/* 1   */
	id = (byte)bdport (PRA1);	/* code */

	DataPortOut ();
	write_code16 (0, 0xf0);

	power_off ();

	if (id & 1) {
		dlp->msgBox (dmfInformation | dmfOKButton,
			"\003Boot Block is locked");
	} else {
		dlp->msgBox (dmfInformation | dmfOKButton,
			"\003Boot Block is not locked");
	}

	return (0);
}


/* ࠢ 稯  ஬. */
int compare_all (void) {

	dword i;
	dword errors = 0;
	dword istep = pchd->size / 64L;
	word dat;

	if (chtype == 0xFFFF)
		return ERC_CHIPUNSP;

	power_on ('V');
	for (i = 0; i < pchd->size; i += 2) {
		dat = read_one (i >> 1);
		if ((dlp->getbytebuf (i) + (dlp->getbytebuf (i+1) << 8)) != dat) {
			/* ⬥ ᮢ */
			dlp->setbmark (i, 1);
			dlp->setbmark (i + 1, 1);
			errors += 2;
		}
		if (i % istep == 0 && dlp->iprogress (i >> 1, 2))
			break;
	}
	power_off ();
	dlp->edata = errors;
	return (0);
}


/* ஢ઠ  . */
int blank_chk (void) {

	dword i;
	dword istep = pchd->size / 64L;

	if (chtype == 0xFFFF)
		return (ERC_CHIPUNSP);

	power_on ('R');
	for (i = 0; i < pchd->size; i += 2) {
		if (read_one (i >> 1) != 0xffff) {
			dlp->adr = i;
			power_off ();
			return (1);
		}
		if (i % istep == 0) {
			if (dlp->iprogress (i >> 1, 2))
				break;
		}
	}
	power_off ();
	return (0);
}


/* ⠭ ⨯ ROM */
word settype (void) {

	word i;

	for (i = 0; i < sizeof(chd) / sizeof(CHDEFS); i++) {
		if (chd[i].size == dlp->esize)
			return (chtype = i);
	}
	return (chtype = 0xFFFF);
}


/* main code */
extern "C" int dmain (void) {

	switch (dlp->funcode) {
	case DFN_LOAD:
		/* ⨢஢ 㯭 㭪樨: */
		dlp->funmask[DFN_FREAD] = 1;
		dlp->funmask[DFN_FWRITE] = 1;
		dlp->funmask[DFN_READID] = 1;
		dlp->funmask[DFN_ERASE] = 1;
		dlp->funmask[DFN_FCOMPARE] = 1;
		dlp->funmask[DFN_FBLANK] = 1;
		dlp->funmask[DFN_TEST] = 0;
		dlp->defstate = 0xFF;
		dlp->copyright = copyright;
		dlp->version = DLLVERSION;
		dlp->cfgkeys = cfgkeys;
		dlp->Vppmax = 140;
		return (0);

	case DFN_INIT:
		if (settype () == 0xFFFF)
			return (ERC_CHIPUNSP);
		pchd = &chd[chtype];
		dlp->dippins = pchd->dpins;
                if (cfg_test)
                	dlp->funmask[DFN_TEST] = 1;
		return (0);

	case DFN_FREAD:
		return (read_all ());

	case DFN_FWRITE:
		return (write_all ());

	case DFN_FCOMPARE:
		return (compare_all ());

	case DFN_FBLANK:
		return (blank_chk ());

	case DFN_READID:
		return (read_id ());

	case DFN_ERASE:
		return (erase ());

	case DFN_TEST:
		return (test ());

	default:
		return (ERC_FUNUNSP);
	}
}


/* E0F */

